import re
import random
from warnings import warn

import numpy as np
from os import environ
environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'  # supress pygame message before import
import pygame
import music21 as m21

# check for most up-to-date version
import muprocdurham.version_check

# see https://github.com/DCMLab/standards/blob/main/harmony.py
dcml_harmony_regex = re.compile("".join(r"""
    ^(\.?
        ((?P<globalkey>[a-gA-G](b*|\#*))\.)?
        ((?P<localkey>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i))\.)?
        ((?P<pedal>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i))\[)?
        (?P<chord>
            (?P<numeral>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i|Ger|It|Fr|@none))
            (?P<form>(%|o|\+|M|\+M))?
            (?P<figbass>(7|65|43|42|2|64|6))?
            (\((?P<changes>((\+|-|\^|v)?(b*|\#*)\d)+)\))?
            (/(?P<relativeroot>((b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i)/?)*))?
        )
        (?P<pedalend>\])?
    )?
    (|(?P<cadence>((HC|PAC|IAC|DC|EC)(\..*?)?)))?
    (?P<phraseend>(\\\\|\{|\}|\}\{))?$
    """.split()))


def seed_everything(seed):
    """Set random seed for reproducibility across multiple libraries."""
    random.seed(seed)  # Python's built-in random module
    np.random.seed(seed)  # NumPy's random module


def show_stream(stream):
    try:
        stream.show()
    except m21.converter.subConverters.SubConverterException:
        warn("Cannot show score, falling back to text representation.", UserWarning)
        stream.show('t')


def play_stream(stream):
    try:
        m21.midi.realtime.StreamPlayer(stream).play()
    except pygame.error:
        warn("Cannot play stream", UserWarning)



#  Copyright (c) 2024 Robert Lieck

import IPython.display
import numpy as np
from scipy.io.wavfile import write as write_wav
import librosa
import matplotlib.pyplot as plt


sampling_rate = 44000


def normalise_wave(wave, max_amp=0.95):
    wave /= abs(wave).max()
    wave *= max_amp


def fade_wave(wave, time=0.01, start=True, end=True):
    # ramp of correct length
    fade_vals = np.linspace(0, 1, int(np.ceil(time * sampling_rate)))
    if start:
        wave[:len(fade_vals)] *= fade_vals
    if end:
        wave[-len(fade_vals):] *= np.flip(fade_vals)


def render(wave, normalise=True, fade=True):
    wave = wave.copy()
    if normalise is True:
        normalise = dict()
    if normalise is not False:
        normalise_wave(wave, **normalise)
    if fade is True:
        fade = dict()
    if fade is not False:
        fade_wave(wave, **fade)
    return wave


def save(wave, file_name, normalise=True, fade=True):
    wave = render(wave=wave, normalise=normalise, fade=fade)
    # convert to 16bit integer
    wave = np.int16(wave * (np.iinfo(np.int16).max - 1))
    write_wav(file_name, sampling_rate, wave)


def load(file):
    x, _ = librosa.load(file, sr=sampling_rate)
    return x


def audio(wave, fade=True):
    wave = render(wave=wave, normalise=False, fade=fade)
    IPython.display.display(IPython.display.Audio(data=wave, rate=sampling_rate))


def audio_add(wave, start_time, audio=None, min_total_time=0):
    offset = int(start_time * sampling_rate)
    # extend audio if required
    required_length = max(offset + len(wave), int(min_total_time * sampling_rate))
    if audio is None or len(audio) < required_length:
        new_audio = np.zeros(required_length)
        if audio is not None:
            new_audio[:len(audio)] = audio
        audio = new_audio
    # add audio
    audio[offset:offset + len(wave)] += wave
    return audio


def sound(func, phases=0., duration=1.):
    # time vector
    time = np.arange(0, duration, 1 / sampling_rate)
    # array of unit-angle steps (corresponding to frequency of 1Hz)
    angle_steps = np.full_like(time, duration * 2 * np.pi / len(time))
    # get frequencies and amplitudes over time
    if callable(func):
        freq_amps = func(time)
    else:
        freq_amps = func
    if not isinstance(freq_amps, tuple):
        freqs = freq_amps
        amps = 1.
    else:
        freqs, amps = freq_amps
    freqs = np.atleast_2d(freqs)
    amps = np.atleast_2d(amps)
    # effective change of angle for different frequencies (time-dependent)
    angle_steps = angle_steps[:, None] * freqs
    # actual angle at given time corresponds to the accumulated angle steps
    angles = np.cumsum(angle_steps, axis=0) + phases
    # generate oscillations, multiply by amplitudes and sum up
    return (amps * np.sin(angles)).sum(axis=1)


def spectrogram(wave, ylim=None, figsize=(14, 5), **kwargs):
    X = librosa.stft(wave)
    Xdb = librosa.amplitude_to_db(abs(X))
    Xdb += Xdb.min()
    plt.figure(figsize=figsize)
    kwargs = {**dict(sr=sampling_rate, x_axis='time', y_axis='hz'),
              **dict(kwargs)}
    librosa.display.specshow(Xdb, **kwargs)
    if ylim is not None:
        plt.ylim(None, ylim)
    plt.show()


def harmonic_tone(f0, decay=1, n=20, **kwargs):
    return sound(lambda time: (
        [f0 * i for i in range(1, n + 1)],
        [np.exp(-i*decay) for i in range(0, n)],
    ), **kwargs)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from pitchtypes import EnharmonicPitchClass, EnharmonicPitch


def bigram_matrix_from_model(n_gram_model, counts=False):
    min_val = min(n_gram_model.alphabet)
    max_val = max(n_gram_model.alphabet)
    n = max_val - min_val + 1
    if counts:
        mat = np.zeros((n, n), dtype=int)
    else:
        mat = np.zeros((n, n))
    for from_idx, from_symbol in enumerate(range(min_val, max_val + 1)):
        for to_idx, to_symbol in enumerate(range(min_val, max_val + 1)):
            if counts:
                mat[from_idx, to_idx] = n_gram_model.c((from_symbol, to_symbol))
            else:
                mat[from_idx, to_idx] = n_gram_model.p((from_symbol, to_symbol))
    return mat, (min_val, max_val)


def show_bigram_matrix(mat, minmax=None, show_values=False, names=False, pitch_classes=False, origin='upper', show=True,
                       key_depth=3, black_key_depth=2):
    assert mat.shape[0] == mat.shape[1], f"Bigram matrix must be square, but has shape {mat.shape}"
    assert np.all(mat >= 0), "Values in 'mat' must be non-negative"
    if origin not in ['upper', 'lower']:
        raise ValueError(f"'origin' has to be 'upper' or 'lower' but provided value was: '{origin}'")
    black_key_depth = min(black_key_depth, key_depth)
    # get basic heatmap
    norm = plt.Normalize(vmin=0)
    img1 = cm.viridis(norm(mat))
    # add piano keys if minmax is provided
    if minmax is None:
        # minmax not provided: use basic heatmap
        extent = np.array([0, mat.shape[0] - 1, 0, mat.shape[1] - 1]) - 0.5
        value_extent = extent
        img2 = img1
    else:
        # minmax IS provided: add row/column with b/w piano keys
        min_val, max_val = minmax
        img2 = np.ones((img1.shape[0] + key_depth, img1.shape[1] + key_depth, img1.shape[2]), dtype=img1.dtype)
        img2[key_depth:, key_depth:] = img1
        idx = np.arange(img1.shape[0]) + key_depth
        pitch = idx + min_val
        black_keys = (((pitch - 5 - key_depth) % 12) * 7) % 12 > 6
        img2[0:black_key_depth, idx[black_keys], 0:3] = 0  # x-axis
        if origin == 'upper':
            img2[idx[black_keys], key_depth - black_key_depth:key_depth, 0:3] = 0  # y-axis
        else:  # 'lower'
            img2[idx[black_keys], 0:black_key_depth, 0:3] = 0  # y-axis
        value_extent = np.array([min_val, max_val + 1, min_val, max_val + 1]) - 0.5
        extent = value_extent - [key_depth, 0, key_depth, 0]
    plt.imshow(img2, origin='lower', extent=extent)
    plt.gca().xaxis.set_major_locator(plt.MaxNLocator(integer=True))
    plt.gca().yaxis.set_major_locator(plt.MaxNLocator(integer=True))
    if origin == 'upper':
        plt.gca().invert_yaxis()
        plt.gca().xaxis.tick_top()
        plt.gca().xaxis.set_label_position('top')
    if show_values:
        for i, x in enumerate(np.linspace(value_extent[0] + 0.5, value_extent[1] - 0.5, mat.shape[0], endpoint=True)):
            for j, y in enumerate(np.linspace(value_extent[2] + 0.5, value_extent[3] - 0.5, mat.shape[1], endpoint=True)):
                plt.text(x, y, str(mat[j, i]), ha='center', va='center')
    plt.ylabel("from")
    plt.xlabel("to")
    if names:
        if pitch_classes:
            def name_func(x):
                return str(EnharmonicPitchClass(x))
        else:
            def name_func(x):
                return str(EnharmonicPitch(x))
        plt.gca().xaxis.set_major_formatter(lambda x, pos: name_func(x))
        plt.gca().yaxis.set_major_formatter(lambda x, pos: name_func(x))
    plt.colorbar(cm.ScalarMappable(norm=norm), ax=plt.gca())
    if show:
        plt.show()


def show_unigram_distribution(n_gram_model, counts=False, sort=False, show=True, names=False, pitch_classes=False):
    k = len(n_gram_model.alphabet)
    p = np.zeros(k)
    alphabet = n_gram_model.alphabet
    if sort:
        alphabet = sorted(alphabet)
    for i, a in enumerate(alphabet) :
        if counts:
            p[i] = n_gram_model.c((a,))
        else:
            p[i] = n_gram_model.p((a,))
    plt.bar(np.arange(k), p)
    if names:
        if pitch_classes:
            alphabet = [EnharmonicPitchClass(a) for a in alphabet]
        else:
            alphabet = [EnharmonicPitch(a) for a in alphabet]
    plt.xticks(np.arange(k), alphabet)
    if show:
        plt.show()



from triangularmap import DictTMap, TMap
from collections import defaultdict


def cat_pretty(trees, off=False, *args, **kwargs):
    if off:
        return '\n'.join([tree.pretty(*args, **kwargs) for tree in trees])
    else:
        return '\n'.join(
            list('|'.join(x) for x in zip(*[tree.pretty(*args, **kwargs).split('\n') for tree in trees]))
        )


class PrettySet(set):
    def __init__(self, *args, empty="∅", compact=True, left="{", right="}", **kwargs):
        super().__init__(*args, **kwargs)
        self.empty = empty
        self.compact = compact
        self.left = left
        self.right = right

    def __str__(self):
        if self:
            if self.compact:
                return self.left + ",".join([str(x) for x in self]) + self.right
            else:
                return str(set(self))
        else:
            return self.empty

    def __repr__(self):
        return self.__str__()


class PrettyDict(defaultdict):
    def __init__(self, *args, empty="{}", compact=True, left="{", right="}", **kwargs):
        super().__init__(*args, **kwargs)
        self.empty = empty
        self.compact = compact
        self.left = left
        self.right = right

    def __str__(self):
        if self:
            if self.compact:
                return self.left + ",".join([f"{k}:{v}" for k, v in self.items()]) + self.right
            else:
                return str(dict(self))
        else:
            return self.empty


class SetChart(TMap):
    @classmethod
    def from_dict_set_chart(cls, chart, **kwargs):
        new_chart = SetChart(chart.n, **kwargs)
        for k, v in chart.arr.items():
            for vk in v.keys():
                new_chart.arr[k].add(vk)
        return new_chart

    def __new__(cls, n, **kwargs):
        return DictTMap(n, lambda: PrettySet(**kwargs))


class DictSetChart(TMap):
    def __new__(cls, n, **kwargs):
        return DictTMap(n, lambda: PrettyDict(PrettySet))


from IPython.core.display import display, HTML


def no_linebreaks():
    display(HTML("<style>div.output_area pre {white-space: pre;}</style>"))  # jupyter notebook
    display(HTML("<style>div.jp-OutputArea-output pre {white-space: pre;}</style>"))  # jupyter lab
